@@ -1,28 +1,16 @@
 """Contains various object definitions needed by the weather utility."""
 
 weather_copyright = """\
-# Copyright (c) 2006-2023 Jeremy Stanley <fungi@yuggoth.org>. Permission to
+# Copyright (c) 2006-2024 Jeremy Stanley <fungi@yuggoth.org>. Permission to
 # use, copy, modify, and distribute this software is granted under terms
 # provided in the LICENSE file distributed with this software.
 #"""
 
-weather_version = "2.4.4"
+weather_version = "2.5.0"
 
 radian_to_km = 6372.795484
 radian_to_mi = 3959.871528
 
-def pyversion(ref=None):
-    """Determine the Python version and optionally compare to a reference."""
-    import platform
-    ver = platform.python_version()
-    if ref:
-        return [
-            int(x) for x in ver.split(".")[:2]
-        ] >= [
-            int(x) for x in ref.split(".")[:2]
-        ]
-    else: return ver
-
 class Selections:
     """An object to contain selection data."""
     def __init__(self):
@@ -130,7 +118,7 @@ def filter_units(line, units="imperial"):
     # filter lines with both pressures in the form of "X inches (Y hPa)" or
     # "X in. Hg (Y hPa)"
     dual_p = re.match(
-        "(.* )(\d*(\.\d+)? (inches|in\. Hg)) \((\d*(\.\d+)? hPa)\)(.*)",
+        r"(.* )(\d*(\.\d+)? (inches|in\. Hg)) \((\d*(\.\d+)? hPa)\)(.*)",
         line
     )
     if dual_p:
@@ -139,7 +127,7 @@ def filter_units(line, units="imperial"):
         elif units == "metric": line = preamble + hpa + trailer
     # filter lines with both temperatures in the form of "X F (Y C)"
     dual_t = re.match(
-        "(.* )(-?\d*(\.\d+)? F) \((-?\d*(\.\d+)? C)\)(.*)",
+        r"(.* )(-?\d*(\.\d+)? F) \((-?\d*(\.\d+)? C)\)(.*)",
         line
     )
     if dual_t:
@@ -150,7 +138,7 @@ def filter_units(line, units="imperial"):
     # "Y kilometer(s)"
     if units == "metric":
         imperial_d = re.match(
-            "(.* )(\d+)( mile\(s\))(.*)",
+            r"(.* )(\d+)( mile\(s\))(.*)",
             line
         )
         if imperial_d:
@@ -160,7 +148,7 @@ def filter_units(line, units="imperial"):
     # filter speeds in the form of "X MPH (Y KT)" to just "X MPH"; if metric is
     # desired, convert to "Z KPH"
     imperial_s = re.match(
-        "(.* )(\d+)( MPH)( \(\d+ KT\))(.*)",
+        r"(.* )(\d+)( MPH)( \(\d+ KT\))(.*)",
         line
     )
     if imperial_s:
@@ -170,7 +158,7 @@ def filter_units(line, units="imperial"):
             line = preamble + str(int(round(int(mph)*1.609344))) + " KPH" + \
                 trailer
     imperial_s = re.match(
-        "(.* )(\d+)( MPH)( \(\d+ KT\))(.*)",
+        r"(.* )(\d+)( MPH)( \(\d+ KT\))(.*)",
         line
     )
     if imperial_s:
@@ -182,7 +170,7 @@ def filter_units(line, units="imperial"):
     # if imperial is desired, qualify given forcast temperatures like "X F"; if
     # metric is desired, convert to "Y C"
     imperial_t = re.match(
-        "(.* )(High |high |Low |low )(\d+)(\.|,)(.*)",
+        r"(.* )(High |high |Low |low )(\d+)(\.|,)(.*)",
         line
     )
     if imperial_t:
@@ -204,15 +192,7 @@ def get_uri(
     cachedir="."
 ):
     """Return a string containing the results of a URI GET."""
-    if pyversion("3"):
-        import urllib, urllib.error, urllib.request
-        URLError = urllib.error.URLError
-        urlopen = urllib.request.urlopen
-    else:
-        import urllib2 as urllib
-        URLError = urllib.URLError
-        urlopen = urllib.urlopen
-    import os, time
+    import os, time, urllib, urllib.error, urllib.request
     if cache_data:
         dcachedir = os.path.join( os.path.expanduser(cachedir), "datacache" )
         if not os.path.exists(dcachedir):
@@ -230,8 +210,8 @@ def get_uri(
         dcache_fd.close()
     else:
         try:
-            data = urlopen(uri).read().decode("utf-8")
-        except URLError:
+            data = urllib.request.urlopen(uri).read().decode("utf-8")
+        except urllib.error.URLError:
             if ignore_fail: return ""
             import os, sys
             sys.stderr.write("%s error: failed to retrieve\n   %s\n\n" % (
@@ -273,7 +253,7 @@ def get_metar(
         cacheage=cacheage,
         cachedir=cachedir
     )
-    if pyversion("3") and type(metar) is bytes: metar = metar.decode("utf-8")
+    if type(metar) is bytes: metar = metar.decode("utf-8")
     if verbose: return metar
     else:
         import re
@@ -320,7 +300,8 @@ def get_alert(
     quiet=False,
     cache_data=False,
     cacheage=900,
-    cachedir="."
+    cachedir=".",
+    delay=1
 ):
     """Return alert notice for the specified URI."""
     if not uri:
@@ -332,23 +313,35 @@ def get_alert(
         cacheage=cacheage,
         cachedir=cachedir
     ).strip()
-    if pyversion("3") and type(alert) is bytes: alert = alert.decode("utf-8")
+    if type(alert) is bytes: alert = alert.decode("utf-8")
     if alert:
         if verbose: return alert
         else:
-            if alert.find("\nNATIONAL WEATHER SERVICE") == -1:
-                muted = False
-            else:
+            import re
+            if re.search(r"\nNational Weather Service", alert):
                 muted = True
+            else:
+                muted = False
+            expirycheck = re.search(r"Expires:([0-9]{12})", alert)
+            if expirycheck:
+                # only report alerts and forecasts that expired less than delay
+                # hours ago
+                import datetime, zoneinfo
+                expiration = datetime.datetime.fromisoformat(
+                    "%s-%s-%sT%s:%s" % (
+                        expirycheck[1][:4],
+                        expirycheck[1][4:6],
+                        expirycheck[1][6:8],
+                        expirycheck[1][8:10],
+                        expirycheck[1][-2:],
+                    )).replace(tzinfo=zoneinfo.ZoneInfo("UTC"))
+                now = datetime.datetime.now(tz=zoneinfo.ZoneInfo("UTC"))
+                if now - expiration > datetime.timedelta(hours=delay):
+                    return ""
             lines = alert.split("\n")
-            import time
-            valid_time = time.strftime("%Y%m%d%H%M")
             output = []
             for line in lines:
-                if line.startswith("Expires:") \
-                    and "Expires:" + valid_time > line:
-                    return ""
-                if muted and line.startswith("NATIONAL WEATHER SERVICE"):
+                if muted and line.startswith("National Weather Service"):
                     muted = False
                     line = ""
                 elif line == "&&":
@@ -394,11 +387,11 @@ def get_options(config):
             + "flash_flood_statement," \
             + "flash_flood_warning," \
             + "flash_flood_watch," \
-            + "flood_statement," \
             + "flood_warning," \
             + "severe_thunderstorm_warning," \
             + "severe_weather_statement," \
             + "special_weather_statement," \
+            + "tornado," \
             + "urgent_weather_message"
     option_parser.add_option("--atypes",
         dest="atypes",
@@ -430,6 +423,15 @@ def get_options(config):
         default=default_cachedir,
         help="directory for storing cached searches and data")
 
+    # the --delay option
+    if config.has_option("default", "delay"):
+        default_delay = config.getint("default", "delay")
+    else: default_delay = 1
+    option_parser.add_option("--delay",
+        dest="delay",
+        default=default_delay,
+        help="hours to delay alert and forecast expiration")
+
     # the -f/--forecast option
     if config.has_option("default", "forecast"):
         default_forecast = config.getboolean("default", "forecast")
@@ -596,10 +598,8 @@ def get_options(config):
 
 def get_config():
     """Parse the aliases and configuration."""
-    if pyversion("3"): import configparser
-    else: import ConfigParser as configparser
+    import configparser, os
     config = configparser.ConfigParser()
-    import os
     rcfiles = [
         "/etc/weatherrc",
         "/etc/weather/weatherrc",
@@ -609,10 +609,7 @@ def get_config():
         ]
     for rcfile in rcfiles:
         if os.access(rcfile, os.R_OK):
-            if pyversion("3"):
-                config.read(rcfile, encoding="utf-8")
-            else:
-                config.read(rcfile)
+            config.read(rcfile, encoding="utf-8")
     for section in config.sections():
         if section != section.lower():
             if config.has_section(section.lower()):
@@ -624,9 +621,7 @@ def get_config():
 
 def integrate_search_cache(config, cachedir, setpath):
     """Add cached search results into the configuration."""
-    if pyversion("3"): import configparser
-    else: import ConfigParser as configparser
-    import os, time
+    import configparser, os, time
     scache_fn = os.path.join( os.path.expanduser(cachedir), "searches" )
     if not os.access(scache_fn, os.R_OK): return config
     scache_fd = open(scache_fn)
@@ -648,10 +643,7 @@ def integrate_search_cache(config, cachedir, setpath):
             pass
         return config
     scache = configparser.ConfigParser()
-    if pyversion("3"):
-        scache.read(scache_fn, encoding="utf-8")
-    else:
-        scache.read(scache_fn)
+    scache.read(scache_fn, encoding="utf-8")
     for section in scache.sections():
         if not config.has_section(section):
             config.add_section(section)
@@ -707,9 +699,7 @@ def guess(
     quiet=False
 ):
     """Find URIs using airport, gecos, placename, station, ZCTA/ZIP, zone."""
-    import codecs, datetime, time, os, re, sys
-    if pyversion("3"): import configparser
-    else: import ConfigParser as configparser
+    import codecs, configparser, datetime, time, os, re, sys
     datafiles = data_index(path)
     if re.match("[A-Za-z]{3}$", expression): searchtype = "airport"
     elif re.match("[A-Za-z0-9]{4}$", expression): searchtype = "station"
@@ -744,15 +734,9 @@ def guess(
         datafile = datafiles[dataname][0]
         if datafile.endswith(".gz"):
             import gzip
-            if pyversion("3"):
-                stations.read_string(
-                    gzip.open(datafile).read().decode("utf-8") )
-            else: stations.readfp( gzip.open(datafile) )
+            stations.read_string( gzip.open(datafile).read().decode("utf-8") )
         else:
-            if pyversion("3"):
-                stations.read(datafile, encoding="utf-8")
-            else:
-                stations.read(datafile)
+            stations.read(datafile, encoding="utf-8")
     else:
         message = "%s error: can't find \"%s\" data file\n" % (
             os.path.basename( sys.argv[0] ),
@@ -766,14 +750,9 @@ def guess(
         datafile = datafiles[dataname][0]
         if datafile.endswith(".gz"):
             import gzip
-            if pyversion("3"):
-                zones.read_string( gzip.open(datafile).read().decode("utf-8") )
-            else: zones.readfp( gzip.open(datafile) )
+            zones.read_string( gzip.open(datafile).read().decode("utf-8") )
         else:
-            if pyversion("3"):
-                zones.read(datafile, encoding="utf-8")
-            else:
-                zones.read(datafile)
+            zones.read(datafile, encoding="utf-8")
     else:
         message = "%s error: can't find \"%s\" data file\n" % (
             os.path.basename( sys.argv[0] ),
@@ -795,15 +774,10 @@ def guess(
             datafile = datafiles[dataname][0]
             if datafile.endswith(".gz"):
                 import gzip
-                if pyversion("3"):
-                    airports.read_string(
-                        gzip.open(datafile).read().decode("utf-8") )
-                else: airports.readfp( gzip.open(datafile) )
+                airports.read_string(
+                    gzip.open(datafile).read().decode("utf-8") )
             else:
-                if pyversion("3"):
-                    airports.read(datafile, encoding="utf-8")
-                else:
-                    airports.read(datafile)
+                airports.read(datafile, encoding="utf-8")
         else:
             message = "%s error: can't find \"%s\" data file\n" % (
                 os.path.basename( sys.argv[0] ),
@@ -887,15 +861,9 @@ def guess(
             datafile = datafiles[dataname][0]
             if datafile.endswith(".gz"):
                 import gzip
-                if pyversion("3"):
-                    zctas.read_string(
-                        gzip.open(datafile).read().decode("utf-8") )
-                else: zctas.readfp( gzip.open(datafile) )
+                zctas.read_string( gzip.open(datafile).read().decode("utf-8") )
             else:
-                if pyversion("3"):
-                    zctas.read(datafile, encoding="utf-8")
-                else:
-                    zctas.read(datafile)
+                zctas.read(datafile, encoding="utf-8")
         else:
             message = "%s error: can't find \"%s\" data file\n" % (
                 os.path.basename( sys.argv[0] ),
@@ -948,15 +916,9 @@ def guess(
             datafile = datafiles[dataname][0]
             if datafile.endswith(".gz"):
                 import gzip
-                if pyversion("3"):
-                    places.read_string(
-                        gzip.open(datafile).read().decode("utf-8") )
-                else: places.readfp( gzip.open(datafile) )
+                places.read_string( gzip.open(datafile).read().decode("utf-8") )
             else:
-                if pyversion("3"):
-                    places.read(datafile, encoding="utf-8")
-                else:
-                    places.read(datafile)
+                places.read(datafile, encoding="utf-8")
         else:
             message = "%s error: can't find \"%s\" data file\n" % (
                 os.path.basename( sys.argv[0] ),
@@ -1173,10 +1135,7 @@ def guess(
             )
         try:
             scache_existing = configparser.ConfigParser()
-            if pyversion("3"):
-                scache_existing.read(scache_fn, encoding="utf-8")
-            else:
-                scache_existing.read(scache_fn)
+            scache_existing.read(scache_fn, encoding="utf-8")
             if not scache_existing.has_section(search[0]):
                 scache_fd = codecs.open(scache_fn, "a", "utf-8")
                 scache_fd.writelines(search_cache)
@@ -1227,9 +1186,8 @@ def gecos(formatted):
     return tuple(coordinates)
 
 def correlate():
-    import codecs, csv, datetime, hashlib, os, re, sys, time, zipfile
-    if pyversion("3"): import configparser
-    else: import ConfigParser as configparser
+    import codecs, configparser, csv, datetime, hashlib, os, re, sys, time
+    import zipfile, zoneinfo
     for filename in os.listdir("."):
         if re.match("[0-9]{4}_Gaz_counties_national.zip$", filename):
             gcounties_an = filename
@@ -1518,6 +1476,9 @@ def correlate():
             zone = "z".join( fields[:2] ).lower()
             if zone in zones:
                 state = fields[0]
+                description = fields[3].strip()
+                fips = "fips%s"%fields[6]
+                countycode = "%sc%s" % (state.lower(), fips[-3:])
                 if state:
                     zones[zone]["coastal_flood_statement"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
@@ -1525,27 +1486,25 @@ def correlate():
                     zones[zone]["flash_flood_statement"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
                         "flash_flood/statement/%s/%s.txt"
-                        % (state.lower(), zone))
+                        % (state.lower(), countycode))
                     zones[zone]["flash_flood_warning"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
                         "flash_flood/warning/%s/%s.txt"
-                        % (state.lower(), zone))
+                        % (state.lower(), countycode))
                     zones[zone]["flash_flood_watch"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
                         "flash_flood/watch/%s/%s.txt" % (state.lower(), zone))
-                    zones[zone]["flood_statement"] = (
-                        "https://tgftp.nws.noaa.gov/data/watches_warnings/"
-                        "flood/statement/%s/%s.txt" % (state.lower(), zone))
                     zones[zone]["flood_warning"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
-                        "flood/warning/%s/%s.txt" % (state.lower(), zone))
+                        "flood/warning/%s/%s.txt"
+                        % (state.lower(), countycode))
                     zones[zone]["severe_thunderstorm_warning"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
-                        "thunderstorm/%s/%s.txt" % (state.lower(), zone))
+                        "thunderstorm/%s/%s.txt" % (state.lower(), countycode))
                     zones[zone]["severe_weather_statement"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
                         "severe_weather_stmt/%s/%s.txt"
-                        % (state.lower(), zone))
+                        % (state.lower(), countycode))
                     zones[zone]["short_term_forecast"] = (
                         "https://tgftp.nws.noaa.gov/data/forecasts/nowcast/"
                         "%s/%s.txt" % (state.lower(), zone))
@@ -1556,14 +1515,46 @@ def correlate():
                     zones[zone]["state_forecast"] = (
                         "https://tgftp.nws.noaa.gov/data/forecasts/state/"
                         "%s/%s.txt" % (state.lower(), zone))
+                    zones[zone]["tornado"] = (
+                        "https://tgftp.nws.noaa.gov/data/watches_warnings/"
+                        "tornado/%s/%s.txt" % (state.lower(), countycode))
                     zones[zone]["urgent_weather_message"] = (
                         "https://tgftp.nws.noaa.gov/data/watches_warnings/"
                         "non_precip/%s/%s.txt" % (state.lower(), zone))
                     zones[zone]["zone_forecast"] = (
                         "https://tgftp.nws.noaa.gov/data/forecasts/zone/"
                         "%s/%s.txt" % (state.lower(), zone))
-                description = fields[3].strip()
-                fips = "fips%s"%fields[6]
+                tzcode = fields[7]
+                if tzcode == "A":
+                    zones[zone]["tz"] = "US/Alaska"
+                elif tzcode == "AH":
+                    zones[zone]["tz"] = "US/Aleutian"
+                elif tzcode in ("C", "CE", "CM"):
+                    zones[zone]["tz"] = "US/Central"
+                elif tzcode in ("E", "e"):
+                    zones[zone]["tz"] = "US/Eastern"
+                elif tzcode == "F":
+                    zones[zone]["tz"] = "Pacific/Guadalcanal"
+                elif tzcode == "G":
+                    zones[zone]["tz"] = "Pacific/Guam"
+                elif tzcode == "H":
+                    zones[zone]["tz"] = "US/Hawaii"
+                elif tzcode == "J":
+                    zones[zone]["tz"] = "Japan"
+                elif tzcode == "K":
+                    zones[zone]["tz"] = "Pacific/Kwajalein"
+                elif tzcode in ("M", "MC", "MP"):
+                    zones[zone]["tz"] = "US/Mountain"
+                elif tzcode == "m":
+                    zones[zone]["tz"] = "US/Arizona"
+                elif tzcode == "P":
+                    zones[zone]["tz"] = "US/Pacific"
+                elif tzcode == "S":
+                    zones[zone]["tz"] = "US/Samoa"
+                elif tzcode == "V":
+                    zones[zone]["tz"] = "America/Virgin"
+                else:
+                    zones[zone]["tz"] = ""
                 county = fields[5]
                 if county:
                     if description.endswith(county):
@@ -1605,7 +1596,7 @@ def correlate():
     removed = 0
     changed = 0
     overrides = configparser.ConfigParser()
-    overrides.readfp( codecs.open(overrides_fn, "r", "utf8") )
+    overrides.read_file( codecs.open(overrides_fn, "r", "utf8") )
     overrideslog = []
     for section in overrides.sections():
         addopt = 0
@@ -2028,30 +2019,15 @@ def correlate():
     sys.stdout.write(message)
     sys.stdout.flush()
     airports = configparser.ConfigParser()
-    if pyversion("3"):
-        airports.read(airports_fn, encoding="utf-8")
-    else:
-        airports.read(airports_fn)
+    airports.read(airports_fn, encoding="utf-8")
     places = configparser.ConfigParser()
-    if pyversion("3"):
-        places.read(places_fn, encoding="utf-8")
-    else:
-        places.read(places_fn)
+    places.read(places_fn, encoding="utf-8")
     stations = configparser.ConfigParser()
-    if pyversion("3"):
-        stations.read(stations_fn, encoding="utf-8")
-    else:
-        stations.read(stations_fn)
+    stations.read(stations_fn, encoding="utf-8")
     zctas = configparser.ConfigParser()
-    if pyversion("3"):
-        zctas.read(zctas_fn, encoding="utf-8")
-    else:
-        zctas.read(zctas_fn)
+    zctas.read(zctas_fn, encoding="utf-8")
     zones = configparser.ConfigParser()
-    if pyversion("3"):
-        zones.read(zones_fn, encoding="utf-8")
-    else:
-        zones.read(zones_fn)
+    zones.read(zones_fn, encoding="utf-8")
     qalog = []
     places_nocentroid = 0
     places_nodescription = 0
@@ -2093,6 +2069,7 @@ def correlate():
             zctas_nocentroid += 1
     zones_nocentroid = 0
     zones_nodescription = 0
+    zones_notz = 0
     zones_noforecast = 0
     zones_overlapping = 0
     zonetable = {}
@@ -2120,6 +2097,10 @@ def correlate():
         if not zones.has_option(zone, "description"):
             qalog.append("%s: no description\n" % zone)
             zones_nodescription += 1
+        if not zones.has_option(zone, "tz") or not zones.get(
+                zone, "tz") in zoneinfo.available_timezones():
+            qalog.append("%s: no time zone\n" % zone)
+            zones_notz += 1
         if not zones.has_option(zone, "zone_forecast"):
             qalog.append("%s: no forecast\n" % zone)
             zones_noforecast += 1
@@ -2156,6 +2137,8 @@ def correlate():
             print("   %s zones with no centroid"%zones_nocentroid)
         if zones_nodescription:
             print("   %s zones with no description"%zones_nodescription)
+        if zones_notz:
+            print("   %s zones with no time zone"%zones_notz)
         if zones_noforecast:
             print("   %s zones with no forecast"%zones_noforecast)
         if zones_overlapping:
